"""
draw - Functions for specific ways to illustrate data
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

from functools import wraps
from fishbowl.decorator import setup_axes
from fishbowl.color import next_color


def handle_args(func):
    """ Handle standard arguments for plot style functions
    """
    @wraps(func)
    def default_func(x, y, data=None, **kwargs):
        if data is not None and x in data:
            x = data[x]
        if data is not None and y in data:
            y = data[y]
        ax = kwargs.pop('ax', None)
        if not ax:
            ax = plt.gca()
        return func(x, y, ax, **kwargs)
    return default_func


@handle_args
def line(x, y, ax, **kwargs):
    """ Draw a line connecting discrete points x,y

    Passes kwargs to ax.plot for the line.
    """
    yerr = kwargs.pop('yerr', None)
    lines = ax.plot(x, y, zorder=1, **kwargs)
    if yerr is not None:
        ax.errorbar(x, y,
                    yerr=yerr,
                    marker=None,
                    color=lines[0].get_color(),
                    zorder=1,
                    capsize=0)
    ax.scatter(x, y, marker='o', s=80, color='white', zorder=2)
    ax.scatter(x, y, marker='o', s=10, color=lines[0].get_color(), zorder=3)
    return lines


@handle_args
def bar(labels, heights, ax, **kwargs):
    """ Draw bars with heights and corresponding labels

    Passes kwargs to ax.bar
    """
    width = kwargs.pop('width', 0.36)
    offset = kwargs.pop('offset', 0.0)
    if 'color' in kwargs:
        color = kwargs.pop('color')
    else:
        color = next_color()  # make sure not to advance iterator if not used

    dummy = np.arange(0, len(labels)) + offset
    bars = ax.bar(dummy,
                  heights,
                  width=width,
                  color=color,
                  linewidth=0,
                  **kwargs)
    setup_axes(ax, xticks=dummy + width/2.0, xticklabels=labels)
    ax.yaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.grid(axis='y', color='white', ls='-', lw=1.2)
    ax.set_axisbelow(False)
    ax.tick_params(axis='both', which='both', length=0)
    return bars


@handle_args
def barh(labels, widths, ax, **kwargs):
    """ Draw bars with widths and corresponding labels

    Passes kwargs to ax.barh
    """
    height = kwargs.pop('height', 0.36)
    offset = kwargs.pop('offset', 0.0)
    if 'color' in kwargs:
        color = kwargs.pop('color')
    else:
        color = next_color()  # make sure not to advance iterator if not used

    dummy = np.arange(0, len(labels)) + offset
    bars = ax.barh(dummy,
                   widths,
                   height=height,
                   color=color,
                   linewidth=0,
                   **kwargs)
    setup_axes(ax, yticks=dummy + height/2.0, yticklabels=labels)
    ax.xaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.grid(axis='x', color='white', ls='-', lw=1.2)
    ax.set_axisbelow(False)
    ax.tick_params(axis='both', which='both', length=0)
    return bars



"""
core - Style setup and control tools.

Provides simple tools for setting style that should be used the majority
of the time when creating graphics.

style       - Set style temporarily within a ``with`` statement
set_style   - Globally set style according to provided options
reset_style - Globally reset style to matplotlib defaults
get_style   - Return the current style options dictionary
"""

import matplotlib
from fishbowl import color
from fishbowl import font as ft
from fishbowl import axes as ax

from contextlib import contextmanager
from cycler import cycler


_defaultparams = matplotlib.rcParams.copy()


# ------------------------------------------------------------
# Style Configuration
# ------------------------------------------------------------

def _set_style(options):
    """ Internal implementation of style setting.

    """
    _set_style.current_options = options

    palette = options.pop('color.palette', None)
    if palette:
        options['axes.prop_cycle'] = cycler('color', palette)

    # Remaining options are rcParams
    matplotlib.rcParams.update(options)


_set_style.current_options = _defaultparams.copy()


def reset_style():
    """ Return the style to matplotlib defaults.

    """
    soptions = _defaultparams.copy()
    set_style(style=soptions)


def get_style():
    """ Return a complete style dictionary matching the current style.

    This dictionary is mostly the rc parameters from matplotlib, with
    a few additional options handled by set_style.

    """
    return _set_style.current_options.copy()


def set_style(axes='minimal', palette='goldfish',
              cmap='YlGnBu', font='Inconsolata', style=None):
    """ Set the global style.

    Parameters
    ----------
    axes
        The style option for axes that controls x/y-axis, ticks, grid, etc...
        Accepts names of saved axes styles
    palette
        color palette for consecutive objects
        Accepts names of saved palettes or palettable.palette instances
    cmap
        Name of the cmap to use for 2D plots.
        Accepts names of matplotlib cmaps or palettable.palette instances
    fonts
         Name of the font style to use, typically just a font name
         Accepts names of saved font configurations or names of system fonts.
    style
         A dictionary that contains all style options
         If provided other keywords are ignored

    """

    if style:
        _set_style(style)
        return

    style = get_style()

    # Colors
    style.update(color.palette(palette))
    style.update(color.cmap(cmap))

    # Axes
    style.update(ax.axes(axes))

    # Fonts
    style.update(ft.font(font))

    _set_style(style)


@contextmanager
def style(**kwargs):
    """ Context manager for using style settings temporarily.

    See Also
    --------
    set_style: called with kwargs within the context

    """
    initial_style = get_style()
    set_style(**kwargs)
    yield
    set_style(style=initial_style)



"""
color - Definitions and controls for color palettes

Provides a series of custom qualitative palettes and an interface for saving
(and developing) additional palettes. All functions which accept a palette
support the `palettable` package.

This module is used by core to set colors with the simple tools, but can also
be accessed directly to retrieve and customize palettes.
"""

import itertools
import matplotlib

import numpy as np
import matplotlib.pyplot as plt

from cycler import cycler
from fishbowl.base import loads_from_json, saves_to_json


def next_color():
    """ Next color from the mpl color cycle

    Note that this cycle is independent of the one used within mpl
    """
    if not hasattr(next_color, '_cycle'):
        next_color._cycle = matplotlib.rcParams['axes.prop_cycle']
        next_color._cycle = itertools.cycle(next_color._cycle)
    return next(next_color._cycle)['color']


def cmap(arg):
    """ Configuration for cmap specified by arg

    Parameters
    ----------
    arg
        any cmap known by matplotlib or a palettable.palette instance.
    """
    if hasattr(arg, 'mpl_colormap'):
        matplotlib.cm.register_cmap(arg.name, arg.mpl_colormap)
        return {'image.cmap': arg.name}
    elif isinstance(arg, str):
        return {'image.cmap': arg}
    else:
        raise ValueError('Could not interpret the argument provided for cmap: '
                         + str(arg))


@loads_from_json('fishbowl.palettes.json')
def palette(arg):
    """ Configuration for palette specified by arg

    Parameters
    ----------
    arg
        any name known by fishbowl or a palettable.palette instance.
    """
    if hasattr(arg, 'hex_colors'):
        return {'color.palette': arg.hex_colors}


@saves_to_json('fishbowl.palettes.json')
def save_palette(name, config):
    """ Save a new color palette as name.

    Parameters
    ----------
    name : str
        save name for the palette
    config
        list of hex colors, palettable.palette instance, or
        name known by fishbowl
    """
    return palette(config)


def draw_box_palette(colors, save_name=None, block=10,
                     sep=1, background='white'):
    """ Draw a series of boxes of color in a grid to show a palette.

    Parameters
    ----------
    colors
        a list or list of lists of colors for the boxes
        a list of lists is drawn as a 2D grid
        each entry must be a color understood by matplotlib.
    save_name : str
        if provided, saves plot to save_name
    block : int
        size of the side of each box (pixels)
    sep : int
        separation between each box (pixels)
    background
        color of the background between palettes
    """
    if not hasattr(colors[0], "__iter__"):
        colors = [colors]

    # Calculate pixel sizes
    rows = len(colors)
    cols = len(colors[0])
    side = block + 2 * sep
    image = np.zeros((side * rows, side * cols))

    # Assign a unique, consecutive value to each color block
    for col in range(cols):
        for row in range(rows):
            image[sep + side * row: sep + block + side * row, sep +
                  side * col:sep + block + side * col] = rows * col + row + 1

    # Figure without border or axis, sized to just contain the blocks
    fig = plt.figure(frameon=False, figsize=(cols, rows))
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)

    # Show it as an image, using one color per box value in a cmap
    ax.imshow(image,
              cmap=matplotlib.colors.ListedColormap(
                  [background] + list(itertools.chain(*colors))),
              interpolation="nearest")
    if save_name:
        fig.savefig(save_name, dpi=120)


def draw_sin_palette(colors, save_name=None):
    """ Draw a series of sin waves in each color to show a palette.

    Parameters
    ----------
    colors : array_like
        list of colors, each must be understood by matplotlib
    save_name : str
        if provided, saves plot to save_name
    """
    fig, (ax) = plt.subplots()
    ax.set_prop_cycle(cycler('color', colors))
    n = len(colors)
    x = np.linspace(-5, 5, 100)
    for i in range(1, n + 1):
        ax.plot(x, np.sin(x + i * .5) * (n + 1 - i))
    ax.set_axis_off()
    if save_name:
        fig.savefig(save_name, dpi=300)



"""
axes - Configuration of matplotlib axes
"""
from fishbowl.base import loads_from_json, saves_to_json


@loads_from_json('fishbowl.axes.json')
def axes(name):
    """ Configuration for named axes

    name must be a saved (or default) configuration
    """
    # Only handled through the save/load system
    pass


@saves_to_json('fishbowl.axes.json')
def save_axes(name, config):
    """ Save a new axes style as name.

    Parameters
    ----------
    name
        save name for the axes configuration
    config
        can be a dictionary of params or a named axes style
    """
    if isinstance(config, dict):
        return config
    return axes(config)



import itertools
import pydoc
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.offsetbox as offsetbox

from mpl_toolkits.axes_grid.anchored_artists import AnchoredText
from functools import wraps


def _decorate_all(decorators):
    """ Decorate a function with all decorators

    """
    def decorator(f):
        for d in reversed(decorators):
            f = d(f)
        return f
    return decorator


def _plot_helper(func):
    """ Internally mark a function to be used as a plot helper

    Each function marked this way should accept ax, **kwargs
    """
    _plot_helper._functions.append(func)
    return func


_plot_helper._functions = []


@_plot_helper
def setup_axes(ax, **kwargs):
    """ Setup an axis with passed arguments.

    A shortcut for calls of a bunch of axes member functions.
    Additional kwargs are ignored to facilitate argument passing.

    Parameters
    ----------
    title : str
        Text to be placed on top center of axes.
    xmin : float
        Set minimum of x-axis.
    xmax : float
        Set maximum of x-axis.
    ymin : float
        Set minimum of y-axis.
    ymax : float
        Set maximum of y-axis.
    logx : bool
        Use log scale for x-axis
    logy : bool
        Use log scale for y-axis
    xlabel : str
        Label for x-axis.
    ylabel : str
        Label for y-axis.
    xticks : str
        String holding space-separated tick locations
    xticklabels : str
        String hold space-separated tick labels
    xtickrot : float
        Rotation for tick labels
    yticks : str
        String holding space-separated tick locations
    yticklabels : str
        String hold space-separated tick labels
    ytickrot : float
        Rotation for tick labels
    xpercent : bool
        Format the x ticks as percentages
    ypercent : bool
        Format the y ticks as percentages
    xdollar : bool
        Format the x ticks as dollars
    ydollar : bool
        Format the y ticks as dollars

    """

    # Handle some shortcut syntaxes
    if kwargs.get('xticks') is not None and 'xticklabels' not in kwargs:
        kwargs['xticklabels'] = kwargs['xticks']
    if kwargs.get('yticks') is not None and 'yticklabels' not in kwargs:
        kwargs['yticklabels'] = kwargs['yticks']

    # Handle space separated strings that should be lists
    fields = ['xticks', 'yticks', 'xticklabels', 'yticklabels']
    converts = [float, float, str, str]
    for field, convert in zip(fields, converts):
        if field in kwargs:
            try:
                kwargs[field] = [convert(x) for x in kwargs[field].split()]
            except (AttributeError):
                pass

    if kwargs.get('title') is not None:
        ax.set_title(kwargs['title'])

    if kwargs.get('ymin') is not None:
        ax.set_ylim(bottom=float(kwargs['ymin']))
    if kwargs.get('ymax') is not None:
        ax.set_ylim(top=float(kwargs['ymax']))
    if kwargs.get('ylabel') is not None:
        ax.set_ylabel(kwargs['ylabel'])
    if kwargs.get('yticks') is not None:
        ax.set_yticks(kwargs['yticks'])
    if kwargs.get('yticklabels') is not None:
        if kwargs.get('ytickrot') is not None:
            ax.set_yticklabels(kwargs['yticklabels'],
                               rotation=kwargs['ytickrot'])
        else:
            ax.set_yticklabels(kwargs['yticklabels'])

    if kwargs.get('xmin') is not None:
        ax.set_xlim(left=float(kwargs['xmin']))
    if kwargs.get('xmax') is not None:
        ax.set_xlim(right=float(kwargs['xmax']))

    if kwargs.get('xlabel') is not None:
        ax.set_xlabel(kwargs['xlabel'])
    if kwargs.get('xticks') is not None:
        ax.set_xticks(kwargs['xticks'])
    if kwargs.get('xticklabels') is not None:
        if kwargs.get('xtickrot') is not None:
            ax.set_xticklabels(kwargs['xticklabels'],
                               rotation=kwargs['xtickrot'])
        else:
            # Best guess if 90 rotation required
            labels = kwargs['xticklabels']
            maxw = max([len(label) for label in labels])
            if maxw * len(labels) > 45:
                rotation = 90
            else:
                rotation = 0
            ax.set_xticklabels(labels, rotation=rotation)

    if kwargs.get('logx'):
        ax.set_xscale('log')
    if kwargs.get('logy'):
        ax.set_yscale('log')

    if ax.get_xscale() != 'log':
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
    if ax.get_yscale() != 'log':
        ax.yaxis.set_minor_locator(ticker.AutoMinorLocator())
    # Move spinse to back of zorder
    for spine in ax.spines.values():
        spine.set_zorder(100)

    if kwargs.get('xpercent'):
        ax.xaxis.set_major_formatter(ticker.FuncFormatter('{0:.1%}'.format))
    if kwargs.get('ypercent'):
        ax.yaxis.set_major_formatter(ticker.FuncFormatter('{0:.1%}'.format))
    if kwargs.get('xdollar'):
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(r'\${:.0f}'.format))
    if kwargs.get('ydollar'):
        ax.yaxis.set_major_formatter(ticker.FuncFormatter(r'\${:.0f}'.format))
    return


@_plot_helper
def text(ax, text=None, text_location="upper left", **kwargs):
    """ Place text on an axis at text_location

    Additional kwargs are ignored to facilitate argument passing.

    Parameters
    ----------
    text : str
        The text to put on the axis
    text_location : str
        One of matplotlibs string or integer locations or a tuple of x,y

    """
    if text is None:
        return
    locations = {'best': 0,
                 'upper right': 1,
                 'upper left': 2,
                 'lower left': 3,
                 'lower right': 4,
                 'right': 5,
                 'center left': 6,
                 'center right': 7,
                 'lower center': 8,
                 'upper center': 9,
                 'center': 10}
    bbox = {}
    if text_location in locations:
        text_location = locations[text_location]
    elif text_location in locations.values():
        pass
    # handle tuple string
    elif '(' in text_location:
        text_location = tuple(float(x)
                              for x in text_location[1:-1].split(','))
    elif type(text_location) is tuple:
        # Specify x,y coordinates for upper left corner
        bbox['bbox_to_anchor'] = text_location
        bbox['bbox_transform'] = ax.transAxes
        text_location = None

    text = text.replace(";", "\n")
    anchored = AnchoredText(text,
                            prop=dict(size=18),
                            loc=text_location,
                            frameon=False,
                            **bbox)
    ax.add_artist(anchored)


@_plot_helper
def legend(ax, legend=None, legend_text=None, **kwargs):
    """ Place legend on an axis

    Additional kwargs are ignored to facilitate argument passing.

    Parameters
    ----------
    legend : str
        either a mpl name for the location or tuple coordinates
    legend_text : str
        additional text to place above legend
        semicolons are replaced with newlines

    """
    if legend is None:
        return
    bbox = {}
    # Specify x,y coordinates for upper left corner
    if '(' in legend:
        bbox['bbox_to_anchor'] = tuple(float(x)
                                       for x in legend[1:-1].split(','))
        bbox['bbox_transform'] = ax.transAxes
        legend = 2

    handles, labels = ax.get_legend_handles_labels()
    legend = ax.legend(handles,
                       labels,
                       frameon=False,
                       loc=legend,
                       prop={'size': 18},
                       labelspacing=0.25,
                       ncol=1 if len(labels) < 6 else 2,
                       **bbox)

    # Add extra text above legend.
    if legend_text:
        for sub in legend_text.split(";")[::-1]:
            txt = offsetbox.TextArea(sub, {'size': 18})
            box = legend._legend_box
            box.get_children().insert(0, txt)
            box.set_figure(box.figure)
    return legend


def plot(func):
    try:
        import click
    except ImportError:
        click = None

    if click:
        doc_strings = [f.__doc__ for f in _plot_helper._functions]
        decorators = [click.command()]
        chain = itertools.chain(*(s.split("\n") for s in doc_strings))
        lines1, lines2 = itertools.tee(chain)
        next(lines2, None)
        for line1, line2 in itertools.izip(lines1, lines2):
            if ':' in line1:
                opt, t = [s.strip() for s in line1.split(":")]
                decorators.append(click.option('--' + opt,
                                               type=pydoc.locate(t),
                                               help=line2.strip()))
        decorators.append(wraps(func))
    else:
        decorators = [wraps(func)]

    @_decorate_all(decorators)
    def plotted_func(**kwargs):
        fig, ax = plt.subplots()
        name = func(fig, ax, **kwargs)
        for helper in _plot_helper._functions:
            helper(ax, **kwargs)
        fig.savefig(name)
    return plotted_func



"""
fishbowl - Style options and customization for matplotlib

fishbowl provides a minimalist style for matplotlib plots, including nice
presets for axes, colors, and fonts.

Call set_style() to enable globally or use style() in a with statement.

If you want to make full use of the font features, you need the 'pgf' backend.
The easiest way to do this is usually:

>>> import matplotlib
>>> matplotlib.use('pgf')
>>> import fishbowl
>>> import matplotlib.pyplot as plt

Modules:

core -- Style setup and control tools

"""
__version__ = '0.3.1'

from fishbowl.core import (style, set_style,  # noqa: F401
                           get_style, reset_style)
from fishbowl.decorator import plot  # noqa: F401



"""
base - Utilities for saving/loading configurations.
"""

import os
import json

from functools import wraps


def _config_path(path):
    """ Absolute path to file in configuration directory for module.

    """
    return os.path.realpath(os.path.join(os.getcwd(),
                                         os.path.dirname(__file__),
                                         'config',
                                         path))


def loads_from_json(path):
    """ Decorator to make function first return values saved at json path.

    Assumes the first argument is the dictionary key if intended to load from
    file. The function is assumed to handle all cases where the key is not
    found in json.
    """
    def loads_from_json_dec(func):
        # Assign a json path to the function for reading
        func._json = _config_path(path)

        @wraps(func)
        def json_load_func(*args, **kwargs):
            if os.path.exists(func._json) and len(args) == 1:
                with open(func._json, 'r') as infile:
                    saved_values = json.load(infile)
                if args[0] in saved_values:
                    return saved_values[args[0]]
            val = func(*args, **kwargs)
            if val:
                return val
            raise ValueError('Could not interpret argument "'
                             + str(args[0])
                             + '" and no saved value found.')
        return json_load_func
    return loads_from_json_dec


def saves_to_json(path):
    """ Decorator to create functions which save their outputs to json file.

    Assumes first argument is the dictionary key to save to the file, and
    returns the saved value.
    """
    def saves_to_json_dec(func):
        # Assign a json path to the function for writing
        func._json = _config_path(path)

        @wraps(func)
        def json_save_func(*args, **kwargs):
            value = func(*args, **kwargs)

            if os.path.exists(func._json):
                with open(func._json, 'r') as infile:
                    saved_values = json.load(infile)
            else:
                saved_values = {}
            saved_values[args[0]] = value

            with open(func._json, 'w') as outfile:
                json.dump(saved_values, outfile)
            return value
        return json_save_func
    return saves_to_json_dec



#!/usr/bin/env python

from setuptools import setup, find_packages

requirements = [
    'matplotlib>=2.0.0'
]

setup(
    name='fishbowl',
    version='0.3.1',
    description='Customizable matplotlib style extension',
    author='Bradley Axen',
    author_email='bradley.axen@gmail.com',
    url='https://github.com/baxen/fishbowl',
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Visualization',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='style plotting matplotlib',
    install_requires=requirements,
    package_data={
        'fishbowl': ['config/fishbowl.axes.json',
                     'config/fishbowl.palettes.json']
    }
)



"""
font - Configurations for system fonts through latex
"""

from fishbowl.base import loads_from_json, saves_to_json


@loads_from_json('fishbowl.font.json')
def font(name, size=20):
    """ Return configuration for named font

    If a saved font config is not found, assumes the named font is a system
    font and attempts to use it through pgf and xelatex.
    """
    config = {'font.family': 'serif',
              'font.serif': name,
              'font.size': size}
    return config


@saves_to_json('fishbowl.font.json')
def save_font(name, config):
    """ Save a new font specified by config as name.

    Parameters
    ----------
    name : str
        save name for the font
    config
        a dictionary of params or a named font
    """
    if isinstance(config, dict):
        return config
    return font(config)
